/* $Workfile: AlvisBurst.c $       $Revision: 2 $      $Modtime: 10/26/08 8:16p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    AlvisBurst.C

    A simple example console application demonstrating the use of the Burst
    device with the following boards:
    
      Alvis-4-CSI and Alvis-4M-CSI

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.


    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxT1E1.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"
#include "OtxDsp.h"

#define DEMO_ASP_BURST_RX_FILE_NAME "/tmp/rx.raw"

#define DEMO_DEFAULT_LOG_FILE "AlvisBurst.log"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_BURST_BUF_SIZE (4096*2) // 8k

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   rxFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    coLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_UINT16             nLineLength;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_EVENT   g_hBurstNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hBurstEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nSubsystemId = 0;
OTX_UINT32  g_nNoEventCount = 0;


OTX_UINT32  g_nBurstRxDataAvailable = 0;
OTX_UINT32  g_nBurstTxBufDepleted = 0;
OTX_UINT32  g_nBurstOverRunErrors = 0;
OTX_UINT32  g_nBurstUnderRunErrors = 0;
OTX_UINT32  g_nBurstWriteDataErrors = 0;
OTX_UINT32  g_nBurstReadDataErrors = 0;

FILE       *g_pRxFile = NULL;

static OTX_UINT32 g_cbBurstBufSize = 0;  // Set to a valid value later

static OTX_UINT8 *g_sqBurstDmaBuf1 = NULL;
static OTX_UINT8 *g_sqBurstDmaBuf2 = NULL;

OTX_UINT8 *g_pCurrentTxBuf = g_sqBurstDmaBuf1;
OTX_UINT8 *g_pCurrentRxBuf = g_sqBurstDmaBuf2;

OTX_BOOL    g_bLoopTransceiverMode = OTX_FALSE;
OTX_BOOL    g_bFileTransceiverMode = OTX_FALSE;


struct {
  OTX_HANDLE hBoard;
  OTX_HANDLE hAspBurst;
  OTX_HANDLE hQuadE1T1;
  OTX_HANDLE hLi[4];
} g_hPhysDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event for low priority events
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    // D16V86 apps cannot use event objects
    g_hNotificationEvent = (OTX_EVENT) NULL;      
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); 
  DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); 
  DemoCheckRc(result);

    // Create a notification Event high priority (Burst) events
  result = OtxDrvCreateEvent(&g_hBurstNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    // D16V86 apps cannot use event objects
    g_hBurstNotificationEvent = (OTX_EVENT) NULL;      
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for burst device notifications
  result = OtxDrvCreateEventQueue(128, g_hBurstNotificationEvent, &g_hBurstEventQueue); 
  DemoCheckRc(result);

  // Enable Event Queue for the burst device
  result = OtxDrvEnable(g_hBurstEventQueue, OTX_TASK_SYNC,0); 
  DemoCheckRc(result);

}



/*******************************************************************************

  @func
    Open the Physical Devices within Alvis CSI

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_UINT32 nDeviceId;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    DemoExitQuestion();
    break;
  }

  // Open the DaVinci Nic and the DMP device (hosted by the NIC)
#if 1
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 0, 0, g_hEventQueue, &g_hPhysDevs.hBoard); DemoCheckRc(result);
#else
  result = OtxDrvGetDeviceHandle(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 0, &g_hPhysDevs.hBoard); DemoCheckRc(result);
#endif

  // Open the ASP Burst device
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_BURST, 0, 0, g_hBurstEventQueue, &g_hPhysDevs.hAspBurst); DemoCheckRc(result);

  {
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, &g_cbBurstBufSize); DemoCheckRc(result);
    DemoLogf("Default Burst Buffer Size       = %d\n", g_cbBurstBufSize);

    g_cbBurstBufSize = DEMO_BURST_BUF_SIZE;
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, g_cbBurstBufSize); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, &g_cbBurstBufSize); DemoCheckRc(result);
    DemoLogf("Changed Burst Buffer Size       = %d\n", g_cbBurstBufSize);

    g_sqBurstDmaBuf1 = (OTX_UINT8 *) malloc(g_cbBurstBufSize);
    g_sqBurstDmaBuf2 = (OTX_UINT8 *) malloc(g_cbBurstBufSize);

    g_pCurrentTxBuf = g_sqBurstDmaBuf1;
    g_pCurrentRxBuf = g_sqBurstDmaBuf2;

    if (!g_sqBurstDmaBuf1 || !g_sqBurstDmaBuf2) {
      DemoLogf("ERROR: Failed to allocate burst buffers\n", g_demoEnv.nBoardType);
      DemoExitQuestion();
    }

  }

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_SUBSYSTEM_ID,  &g_nSubsystemId); DemoCheckRc(result);
  DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, g_nSubsystemId);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_QUAD_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < 4; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  }

}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);

  }

}

/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hBoard,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    // Enable Quad E1T1 device
    //DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 device\n"1);
    result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;
  default:
    break;
  }


}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a 
  // very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;

  ui32 = g_cbBurstBufSize;
  
  otxRc = OtxBurstSendData(g_hPhysDevs.hAspBurst, nSourceId, g_pCurrentTxBuf, &ui32); 
  
  if (OTX_SUCCESS(otxRc)) {
    if(ui32 != g_cbBurstBufSize) {
      DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer (%d).\n", nSourceId, ui32);
    }
  } else {
    g_nBurstWriteDataErrors++;
  }

  if (g_bLoopTransceiverMode) {
    // Point to the other buffer
    if (g_pCurrentTxBuf == g_sqBurstDmaBuf1) {
      g_pCurrentTxBuf = g_sqBurstDmaBuf2;
    } else {
      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
    }
  }


}


/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;
  int cbWritten;

  ui32 = g_cbBurstBufSize;
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hAspBurst, nSourceId, g_pCurrentRxBuf, &ui32, &nFrameRefId, &dtFrameRefTime); 

  if (OTX_SUCCESS(otxRc)) {
    if (ui32 != g_cbBurstBufSize) {
      DemoLogf("DemoBurstReceiver[%d]: sizeof(g_sqBurstDmaBuf2) [%d] != ui32 [%d].\n", nSourceId, g_cbBurstBufSize, ui32);
    }
    
    if (g_pRxFile && g_bFileTransceiverMode) {
      cbWritten = fwrite(g_pCurrentRxBuf, 1, g_cbBurstBufSize, g_pRxFile);
      if (cbWritten != (int)g_cbBurstBufSize) {
        DemoLogf("DemoBurstReceiver[%d]: Could not write the expected number of bytes to the file (%d).\n", nSourceId, ui32);
      }
    }
  } else {
    g_nBurstReadDataErrors++;
  }

  if (g_bLoopTransceiverMode) {
    // Point to the other buffer
    if (g_pCurrentRxBuf == g_sqBurstDmaBuf1) {
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;
    } else {
      g_pCurrentRxBuf = g_sqBurstDmaBuf1;
    }
  }

}



#define DEMO_EVENT_DATA_BUF_SIZE 10


/*******************************************************************************
  @func
    Idle Function that check for Events from the Burst device only.
    This thread runs at high priority, since the burst device has to be serviced within
    8 ms (assuming 8k byte ping/pong buffers).
*******************************************************************************/
void  *DemoBurstIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;


// Increase the priority
  struct sched_param param;  // scheduling priority
  int policy;     // scheduling policy

  // Get the current thread id
  pthread_t thread_id = pthread_self();

  //DemoLogf("Get scheduling parameters\n");
  pthread_getschedparam(thread_id, &policy, &param);

  DemoLogf("The thread scheduling parameters indicate priority = %d\n", param.sched_priority);

  // To set the scheduling priority of the thread
  policy = SCHED_RR;     // scheduling policy
  param.sched_priority = 90;
  pthread_setschedparam(thread_id, policy, &param);

  while (1) {
    OtxDrvWaitForSingleEvent(g_hBurstNotificationEvent, OTX_TIME_INFINITY);

    do {

      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hBurstEventQueue, 
                                    sqEventData, 
                                    DEMO_EVENT_DATA_BUF_SIZE, 
                                    &coEventsFetched); 
      DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          if (sqEventData[i].m_eDeviceType == OTX_DEVICE_BURST) {
            // Handle the events from the Burst device here (since they are frequent and need special attention)
            switch (sqEventData[i].m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              g_nBurstRxDataAvailable++;
              DemoBurstReceiver(0);
              break;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              g_nBurstTxBufDepleted++;
              DemoBurstSender(0);
              break;
            case OTX_BURST_EC_BUFFER_OVERRUN:  // Receiver data overrun error - data lost
              g_nBurstOverRunErrors++;
              //DemoPrintEvent(&eventData);
              //DemoLogf("ASP Burst Event: Receiver OTX_BURST_EC_BUFFER_OVERRUN\n");
              break;
            case OTX_BURST_EC_BUFFER_UNDERRUN: // Sender data underrun error
              g_nBurstUnderRunErrors++;
              //DemoLogf("ASP Burst Event: Sender OTX_BURST_EC_BUFFER_UNDERRUN\n");
              //DemoPrintEvent(&eventData);
              break;
            default:
              DemoLogf("DemoBurstIdleFunction eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
                sqEventData[i].m_eDeviceType,
                sqEventData[i].m_nSourceId,
                sqEventData[i].m_nCode);
              break;
            }
          } else {
            DemoLogf("DemoBurstIdleFunction eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
              sqEventData[i].m_eDeviceType,
              sqEventData[i].m_nSourceId,
              sqEventData[i].m_nCode);
          }
        }

      } else {
        g_nNoEventCount++;
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {

      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, 
                                    sqEventData, 
                                    DEMO_EVENT_DATA_BUF_SIZE, 
                                    &coEventsFetched); 
      DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)

        for (i = 0; i < coEventsFetched; i++) {
          // Act on the event data.
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DMP_ALVIS:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    {
      switch (pEventData->m_nCode) {
      case OTX_BRD_EC_CORE2CORE_INTERRUPT:
        {
          DemoLogf("Core-to-Core interrupt\n");  // OTX Internal use
        }
        break;
      default:
        DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
          pEventData->m_eDeviceType,
          pEventData->m_nSourceId,
          pEventData->m_nCode);
        break;
      }
    }
    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
      pEventData->m_eDeviceType,
      pEventData->m_nSourceId,
      pEventData->m_nCode);
    break;
  }


}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_DMP_ALVIS_4_CSI;
  envP->coLi              = 4;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->nLineLength       = 10;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;

  strcpy(envP->rxFileName, DEMO_ASP_BURST_RX_FILE_NAME);

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      //OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'R') {   // Rx data file
          strcpy(envP->rxFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type

          if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /T[D|K]        : Board Type used for the demo, where\n");
  DemoLogf("                    D  = Use Alvis-4-CSI board\n");
  DemoLogf("                    K  = Use Alvis-4M-CSI board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /R<fileName>   : Rx Burst Data file name (default is %s)\n", DEMO_ASP_BURST_RX_FILE_NAME);
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /F[E|S|9]      : T1 Frame Format, where\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]        : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]      : E1 Frame Format, where\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]        : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", 
      OTX_RESULT_CODE(result), 
      OTX_RESULT_SOURCE_ID(result), 
      OTX_RESULT_LINE_NUM(result), 
      msgStr);

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" e/D Enable(e)/Diable(E) the ASP DMA Burst Stats\n");
    DemoLogf(" b/B Start(b)/Stop(B) sending/receiving files \n");
    DemoLogf(" x/X Start(x)/Stop(X) ASP DMA Loop mode (RX->TX)\n");
    DemoLogf(" s   Print ASP DMA Burst Stats\n");
    DemoLogf(" l   Print E1/T1 Line status\n");
    break;


  case 'e':
    {
      OtxDeviceStateE  eDevState;
      //int x;
      //int nLi;
      //int nTs;
      //OTX_UINT8 nByteVal = 0;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      g_bLoopTransceiverMode = OTX_FALSE;
      g_bFileTransceiverMode = OTX_FALSE;

#if 1
      memset(g_pCurrentTxBuf, 0xA5, g_cbBurstBufSize);
#else
      // Send  the highway TS number in the timeslot
      for (x = 0; x < g_cbBurstBufSize; /**/) {
        for (nTs = 0; nTs < 128; nTs++) {
          g_pCurrentTxBuf[x++] = nTs;
        }
      }
#endif


      if (eDevState == OTX_DEVICE_WORKING) {
        DemoLogf("The ASP DMA Burst Device is already enabled. No action\n");
        break;
      }

      DemoLogf("Enabling ASP DMA Burst Device. g_cbBurstBufSize=%d\n", g_cbBurstBufSize);
      result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;

  case 'X':
  case 'B':
  case 'E':
    {
      OtxDeviceStateE  eDevState;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("The ASP DMA Burst Device is not enabled. Can't disable it. No action\n");
        break;
      }

      g_bLoopTransceiverMode = OTX_FALSE;
      g_bFileTransceiverMode = OTX_FALSE;
      result = OtxDrvDisable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

      if (g_pRxFile) {
        OsaThreadSleep(300);  // Give the burst device some time to complete any pending events (if any)
        DemoLogf("Closing %s file\n", g_demoEnv.rxFileName);
        fclose(g_pRxFile);
        g_pRxFile = NULL;
      }

      DemoLogf("Disabled ASP DMA Burst Device.\n");
      DemoLogf("Done.\n");
    }
    break;

  case 'b':
    {
      int x;
      OtxDeviceStateE  eDevState;

      DemoLogf("Opening %s file for writing (g_cbBurstBufSize=%d)\n", g_demoEnv.rxFileName, g_cbBurstBufSize);
      g_pRxFile = fopen(g_demoEnv.rxFileName, "wb");
      if (!g_pRxFile) {
        DemoLogf("ERROR: Failed to open %s file.\n", g_demoEnv.rxFileName);
        break;
      }

      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;

      // Send a gamma pattern, which means that we increase the value by one for each frame (continuously until aborted)
      for (x = 0; x < (int)g_cbBurstBufSize; x++) {
        g_pCurrentTxBuf[x] = (OTX_UINT8) x;
      }
      g_bFileTransceiverMode = OTX_TRUE;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("Enabling ASP DMA Burst Device\n");
        result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L);  DemoCheckRc(result);
      }
    }
    break;

  case 'x':
    {
      OtxDeviceStateE  eDevState;

      DemoLogf("Burst loopback mode (g_cbBurstBufSize=%d)\n", g_cbBurstBufSize);

      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;

      g_bLoopTransceiverMode = OTX_TRUE;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("Enabling ASP DMA Burst Device\n");
        result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L);  DemoCheckRc(result);
      }
    }
    break;




  case 's':
    {
      OTX_UINT32 nVal;

      DemoLogf("ASP DMA Burst Stats\n");
      DemoLogf("TX:\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_STARTS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Start             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Complete          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Ping Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_FULL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Pong Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT, &nVal); DemoCheckRc(result);
      DemoLogf("  Byte Count            = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Error             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS, &nVal); DemoCheckRc(result);
      DemoLogf("  Sync Errors           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS, &nVal); DemoCheckRc(result);
      DemoLogf("  HW Underruns          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER, &nVal); DemoCheckRc(result);
      DemoLogf("  User Underruns        = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_RESERVED0, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved0             = %d\n", nVal);

      DemoLogf("RX:\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Start             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Complete          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Ping Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_FULL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Pong Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT, &nVal); DemoCheckRc(result);
      DemoLogf("  Byte Count            = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Error             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS, &nVal); DemoCheckRc(result);
      DemoLogf("  Sync Errors           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS, &nVal); DemoCheckRc(result);
      DemoLogf("  HW Overruns           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER, &nVal); DemoCheckRc(result);
      DemoLogf("  User Overruns         = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED0, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved0             = %d\n", nVal);

      DemoLogf("Event Stats\n");
      DemoLogf("  Rx Data Available     = %d\n", g_nBurstRxDataAvailable);
      DemoLogf("  Tx Buf Depleted       = %d\n", g_nBurstTxBufDepleted);
      DemoLogf("  Rx Overruns           = %d\n", g_nBurstOverRunErrors);
      DemoLogf("  Tx Underruns          = %d\n", g_nBurstUnderRunErrors);
      DemoLogf("  NoEvent Count         = %d\n", g_nNoEventCount);
      
      DemoLogf("Other Counters\n");
      DemoLogf("  Write Data Errors     = %d\n", g_nBurstWriteDataErrors);
      DemoLogf("  Read Data Errors      = %d\n", g_nBurstReadDataErrors);

    }
    break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;
      OTX_INT32 i;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
      DemoLogf("Frame snapshot of ASP DMA Rx Buf:\n");
      for (i = 0; i < 128; i++) {
        if ((i%8) == 0) DemoLogf("[%.3d]: ", i);
        DemoLogf("%.2X ", g_pCurrentRxBuf[i]);
        if (((i+1)%8) == 0) DemoLogf("\n");
      }
      DemoLogf("\n");
            
    }
    break;


  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("AlvisBurst, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Alvis-CSI ASP Burst Demo program.\n");
  DemoLogf("For Alvis-CSI boards\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, 
                                         OTX_ATTR_LIB_TYPE_STATIC, 
                                         &bIsStaticLinkedLibrary); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_FILE_NAME, 
                                       szLibFileName, 
                                       64); 
    DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", 
             szVersionInfo, 
             szLibFileName, 
             bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();


  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling Physical Devices...\n");
  DemoEnablePhysicalDevices();

  DemoLogf("Setting clocks...\n");
  result = OtxBrdSetClocks(g_hPhysDevs.hBoard, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
  //result = OtxBrdSetClocks(g_hPhysDevs.hBoard, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(result);


  // Read Serial Number from the board
  {
    char szSerialNumber[64];

    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, szSerialNumber, 64); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", szSerialNumber);
  }

  // Read Build Number from the board
  {
    char szBrdBuildNumber[32];

    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 32); DemoCheckRc(result); DemoLogf("Board Build Number: %s\n", szBrdBuildNumber);
  }


  // Main Loop. Loop here and wait for 
  // Events from the driver or commands from the user ....

  OsaThreadStart(DemoBurstIdleFunction);   // This thread runs at a higher priority and handles events from the burst device only.
  OsaThreadStart(DemoIdleFunction);

  //OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

#if 1 // Don't use IDLE character
  {
    int i;
    for (i=0; i<4; i++) {
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 48, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 49, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 50, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 51, 0x00); DemoCheckRc(result);
    }
  }
#endif

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  DemoPrintf("Exit AlvisBurst Demo\n");

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


